/* This software is contributed or developed by KYOCERA Corporation. */
/* (C) 2011 KYOCERA Corporation                                      */

/* Copyright (c) 2009-2010, Code Aurora Forum. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
 */

#include <linux/delay.h>
#include <mach/gpio.h>
#include <mach/vreg.h>
#include "msm_fb.h"

/* registers */
#define ILI9341_REG_POWCTLB       0xCF
#define ILI9341_REG_POWCTL1       0xC0
#define ILI9341_REG_POWCTL2       0xC1
#define ILI9341_REG_POWCTLSEQ     0xED
#define ILI9341_REG_DTCTLA        0xE8
#define ILI9341_REG_DTCTLB        0xEA
#define ILI9341_REG_POWCTLA       0xCB
#define ILI9341_REG_PRCTL         0xF7
#define ILI9341_REG_VCOMCTL1      0xC5
#define ILI9341_REG_VCOMCTL2      0xC7
#define ILI9341_REG_IFCTL         0xF6
#define ILI9341_REG_DFCTL         0xB6
#define ILI9341_REG_MEMACTL       0x36
#define ILI9341_REG_COLMOD        0x3A
#define ILI9341_REG_RGBIFSRCTL    0xB0
#define ILI9341_REG_FRCTL_NOR     0xB1
#define ILI9341_REG_FRCTL_IDLE    0xB2
#define ILI9341_REG_FRCTL_PAR     0xB3
#define ILI9341_REG_EN3G          0xF2
#define ILI9341_REG_SETG          0x26
#define ILI9341_REG_PSETG         0xE0
#define ILI9341_REG_NSETG         0xE1
#define ILI9341_REG_SLPOUT        0x11
#define ILI9341_REG_DISPON        0x29
#define ILI9341_REG_MEMW          0x2C
#define ILI9341_REG_DISPOFF       0x28
#define ILI9341_REG_SLPMOD        0x10
#define ILI9341_REG_RDISPIF       0x0D

#define DEBUG_INFO 

static int lcdc_ili9341_panel_off(struct platform_device *pdev);

static int spi_cs;
static int spi_sclk;
static int spi_sdo;
static int spi_sdi;
static int lcd_reset;

static void ili9341_spi_write(uint8 command, boolean isCommand)
{
	unsigned char i, val = 0;
	
	val = command;

	/* Enable the Chip Select */
	gpio_set_value(spi_cs, 0);
	udelay(1);

	/* Send command/data bit */
	gpio_set_value(spi_sdi, isCommand ? 0 : 1);

	gpio_set_value(spi_sclk, 0);
	udelay(1);
	gpio_set_value(spi_sclk, 1);
	udelay(1);

	/* Clock should be high before entering! */
	for (i = 0; i < 8; i++) {
		/* Drive the data */
		if (val & (0x80>>i))
			gpio_set_value(spi_sdi, 1);
		else
			gpio_set_value(spi_sdi, 0);

		/* Drive the Clk low and then high */
		gpio_set_value(spi_sclk, 0);
		udelay(1);
		gpio_set_value(spi_sclk, 1);
		udelay(1);		
	}

	/* Disable the Chip Select */
	gpio_set_value(spi_cs, 1);
	udelay(2);
}

static void ili9341_write_command(uint8 command)
{
	ili9341_spi_write(command, 1);
}

static void ili9341_write_data(uint8 data)
{
	ili9341_spi_write(data, 0);
}

/* vreg(GP3) ctrl */
static struct vreg *vreg_gp3;
void ili9341_vreg_on( void )
{
  int rc = 0;
  vreg_gp3 = vreg_get(NULL, "gp3");
  rc = vreg_set_level(vreg_gp3, 2800);
  if (!rc)
  {
    rc = vreg_enable(vreg_gp3);
  }
  mdelay(100);//TODO
}
void ili9341_vreg_off( void )
{
  int rc = 0;
  vreg_gp3 = vreg_get(NULL, "gp3");
  rc = vreg_disable(vreg_gp3);
}

struct  ili9341_state_type{
	boolean disp_initialized;
	boolean display_on;
	boolean disp_powered_up;
};

static struct ili9341_state_type ili9341_state = { 0 };
static struct msm_panel_common_pdata *lcdc_ili9341_pdata;



static void spi_init(void)
{
	
#if 1
	/* Setting the Default GPIO's */
	spi_sclk = *(lcdc_ili9341_pdata->gpio_num);
	spi_cs   = *(lcdc_ili9341_pdata->gpio_num + 1);
	spi_sdi  = *(lcdc_ili9341_pdata->gpio_num + 2);
	spi_sdo  = *(lcdc_ili9341_pdata->gpio_num + 3);

	/* Set the output so that we dont disturb the slave device */
	gpio_set_value(spi_sclk, 0);
	gpio_set_value(spi_sdi, 0);

	/* Set the Chip Select De-asserted */
	gpio_set_value(spi_cs, 1);
#endif
}

static void ili9341_disp_powerup(void)
{
	if (!ili9341_state.disp_powered_up && !ili9341_state.display_on) {
#if 0		
		/* LCD_RESET sequence */
		gpio_set_value(lcd_reset,1);
		mdelay(50);
		gpio_set_value(lcd_reset,0);
		mdelay(50);
		gpio_set_value(lcd_reset,1);
		mdelay(120);
#ifdef DEBUG_INFO
		printk(KERN_INFO"%s",__func__);
#endif
		ili9341_write_command(0xCF);  
		ili9341_write_data(0x00); 
		//ili9341_write_data(0xD9); 
		ili9341_write_data(0x81); // Illitek FAE 2011/07/18
		ili9341_write_data(0X30); 
 
		ili9341_write_command(0xED);  
		ili9341_write_data(0x64); 
		ili9341_write_data(0x03); 
		ili9341_write_data(0X12); 
		ili9341_write_data(0X81); 
 
		ili9341_write_command(0xE8);  
		ili9341_write_data(0x85); 
		//ili9341_write_data(0x01); 
		ili9341_write_data(0x10); // Illitek FAE 2011/07/18
		ili9341_write_data(0x78);

		ili9341_write_command(0xCB);  
		ili9341_write_data(0x39); 
		ili9341_write_data(0x2C); 
		ili9341_write_data(0x00); 
		ili9341_write_data(0x34); 
		ili9341_write_data(0x02); 
 
		ili9341_write_command(0xF7);  
		ili9341_write_data(0x20); 
 
		ili9341_write_command(0xEA);  
		ili9341_write_data(0x00); 
		ili9341_write_data(0x00); 
 
		ili9341_write_command(0xC0);       //Power control 
		ili9341_write_data(0x26);     //VRH[5:0] 
 
		ili9341_write_command(0xC1);       //Power control 
		ili9341_write_data(0x11);     //SAP[2:0];BT[3:0] 
 
		ili9341_write_command(0xC5);       //VCM control 
		ili9341_write_data(0x2B); 
		ili9341_write_data(0x2C); 
 
		ili9341_write_command(0xC7);       //VCM control2 
		ili9341_write_data(0xD4); 

		ili9341_write_command(0xF6);       
		ili9341_write_data(0x01); 
		ili9341_write_data(0x30); 
		ili9341_write_data(0x06);   //00

		ili9341_write_command(0xB6);       
		ili9341_write_data(0x0A); 
		ili9341_write_data(0xc2);   //a2
 
		ili9341_write_command(0x36);       // Memory Access Control 
		//ili9341_write_data(0x48); 
		ili9341_write_data(0x08); // Illitek FAE 2011/07/18

		ili9341_write_command(0x3A);       
		ili9341_write_data(0x60);   //55

		ili9341_write_command(0xB0);    
		ili9341_write_data(0xc0); 

		ili9341_write_command(0xB1);    
		ili9341_write_data(0x00); 
		ili9341_write_data(0x19); 

		ili9341_write_command(0xF2);       // 3Gamma Function Disable 
		ili9341_write_data(0x00); 
 
		ili9341_write_command(0x26);       //Gamma curve selected 
		ili9341_write_data(0x01); 
 
		ili9341_write_command(0xE0);       //Set Gamma 
		ili9341_write_data(0x0F); 
		ili9341_write_data(0x25); 
		ili9341_write_data(0x23); 
		ili9341_write_data(0x0A); 
		ili9341_write_data(0x0E); 
		ili9341_write_data(0x09); 
		ili9341_write_data(0x50); 
		ili9341_write_data(0xA9); 
		ili9341_write_data(0x41); 
		ili9341_write_data(0x07); 
		ili9341_write_data(0x10); 
		ili9341_write_data(0x03); 
		ili9341_write_data(0x28); 
		ili9341_write_data(0x2A); 
		ili9341_write_data(0x00); 
 
		ili9341_write_command(0XE1);       //Set Gamma 
		ili9341_write_data(0x00); 
		ili9341_write_data(0x1A); 
		ili9341_write_data(0x1C); 
		ili9341_write_data(0x05); 
		ili9341_write_data(0x11); 
		ili9341_write_data(0x06); 
		ili9341_write_data(0x2F); 
		ili9341_write_data(0x56); 
		ili9341_write_data(0x3E); 
		ili9341_write_data(0x08); 
		ili9341_write_data(0x0F); 
		ili9341_write_data(0x0C); 
		ili9341_write_data(0x17); 
		ili9341_write_data(0x15); 
		ili9341_write_data(0x0F);
		mdelay(50);
		ili9341_write_command(0x11);       //Exit Sleep 
		mdelay(120); 
		ili9341_write_command(0x29);       //Display on 
		
#endif
		ili9341_state.disp_powered_up = TRUE;
	}
}

static void ili9341_disp_on(void)
{
	if (ili9341_state.disp_powered_up && !ili9341_state.display_on) {
		/* LCD_RESET sequence */
		lcdc_ili9341_pdata->panel_config_gpio(1);
		lcd_reset = *(lcdc_ili9341_pdata->gpio_num + 4);
		gpio_set_value(lcd_reset,1);
		mdelay(50);
		gpio_set_value(lcd_reset,0);
		mdelay(50);
		gpio_set_value(lcd_reset,1);
		mdelay(120);
#ifdef DEBUG_INFO
		printk(KERN_INFO"%s",__func__);
#endif
		ili9341_write_command(0xCF);  
		ili9341_write_data(0x00); 
		//ili9341_write_data(0xD9); 
		ili9341_write_data(0x81); // Illitek FAE 2011/07/18
		ili9341_write_data(0X30); 
 
		ili9341_write_command(0xED);  
		ili9341_write_data(0x64); 
		ili9341_write_data(0x03); 
		ili9341_write_data(0X12); 
		ili9341_write_data(0X81); 
 
		ili9341_write_command(0xE8);  
		ili9341_write_data(0x85); 
		//ili9341_write_data(0x01); 
		ili9341_write_data(0x10); // Illitek FAE 2011/07/18
		ili9341_write_data(0x78);

		ili9341_write_command(0xCB);  
		ili9341_write_data(0x39); 
		ili9341_write_data(0x2C); 
		ili9341_write_data(0x00); 
		ili9341_write_data(0x34); 
		ili9341_write_data(0x02); 
 
		ili9341_write_command(0xF7);  
		ili9341_write_data(0x20); 
 
		ili9341_write_command(0xEA);  
		ili9341_write_data(0x00); 
		ili9341_write_data(0x00); 
 
		ili9341_write_command(0xC0);       //Power control 
		ili9341_write_data(0x26);     //VRH[5:0] 
 
		ili9341_write_command(0xC1);       //Power control 
		ili9341_write_data(0x11);     //SAP[2:0];BT[3:0] 
 
		ili9341_write_command(0xC5);       //VCM control 
		ili9341_write_data(0x2B); 
		ili9341_write_data(0x2C); 
 
		ili9341_write_command(0xC7);       //VCM control2 
		ili9341_write_data(0xD4); 

		ili9341_write_command(0xF6);       
		ili9341_write_data(0x01); 
		ili9341_write_data(0x30); 
		ili9341_write_data(0x06);   //00

		ili9341_write_command(0xB6);       
		ili9341_write_data(0x0A); 
		ili9341_write_data(0xc2);   //a2
 
		ili9341_write_command(0x36);       // Memory Access Control 
		//ili9341_write_data(0x48); 
		ili9341_write_data(0x08); // Illitek FAE 2011/07/18

		ili9341_write_command(0x3A);       
		ili9341_write_data(0x60);   //55

		ili9341_write_command(0xB0);    
		ili9341_write_data(0xc0); 

		ili9341_write_command(0xB1);    
		ili9341_write_data(0x00); 
		ili9341_write_data(0x19); 

		ili9341_write_command(0xF2);       // 3Gamma Function Disable 
		ili9341_write_data(0x00); 
 
		ili9341_write_command(0x26);       //Gamma curve selected 
		ili9341_write_data(0x01); 
 
		ili9341_write_command(0xE0);       //Set Gamma 
		ili9341_write_data(0x0F); 
		ili9341_write_data(0x25); 
		ili9341_write_data(0x23); 
		ili9341_write_data(0x0A); 
		ili9341_write_data(0x0E); 
		ili9341_write_data(0x09); 
		ili9341_write_data(0x50); 
		ili9341_write_data(0xA9); 
		ili9341_write_data(0x41); 
		ili9341_write_data(0x07); 
		ili9341_write_data(0x10); 
		ili9341_write_data(0x03); 
		ili9341_write_data(0x28); 
		ili9341_write_data(0x2A); 
		ili9341_write_data(0x00); 
 
		ili9341_write_command(0XE1);       //Set Gamma 
		ili9341_write_data(0x00); 
		ili9341_write_data(0x1A); 
		ili9341_write_data(0x1C); 
		ili9341_write_data(0x05); 
		ili9341_write_data(0x11); 
		ili9341_write_data(0x06); 
		ili9341_write_data(0x2F); 
		ili9341_write_data(0x56); 
		ili9341_write_data(0x3E); 
		ili9341_write_data(0x08); 
		ili9341_write_data(0x0F); 
		ili9341_write_data(0x0C); 
		ili9341_write_data(0x17); 
		ili9341_write_data(0x15); 
		ili9341_write_data(0x0F);
		mdelay(50);
		ili9341_write_command(0x11);       //Exit Sleep 
		mdelay(120); 
		ili9341_write_command(0x29);       //Display on 
		
#if 0		
		mdelay(50);
		ili9341_write_command(0x11);       //Exit Sleep 
		mdelay(120); 
		ili9341_write_command(0x29);       //Display on
		mdelay(100);//TODO
		ili9341_write_command(0x2A);
		ili9341_write_data(0x00);
		ili9341_write_data(0x00);
		ili9341_write_data(0x00);
		ili9341_write_data(239);// 0->239
		ili9341_write_command(0x2B);
		ili9341_write_data(0x00);
		ili9341_write_data(0x00);
		ili9341_write_data(319>>8);
		ili9341_write_data(319&0xff);// 0->319
		ili9341_write_command(0x2C);       //GRAM write
		mdelay(100);//TODO
#endif

		ili9341_state.display_on = TRUE;
	}
}

static int lcdc_ili9341_panel_on(struct platform_device *pdev)
{

	if (!ili9341_state.disp_initialized) {
#if 0
		ili9341_vreg_on();
		/* Configure reset GPIO that drives DAC */

		lcdc_ili9341_pdata->panel_config_gpio(1);
		lcd_reset = *(lcdc_ili9341_pdata->gpio_num + 4);
		/* LCD_RESET sequence */
		gpio_set_value(lcd_reset,1);
		mdelay(50);
		gpio_set_value(lcd_reset,0);
		mdelay(50);
		gpio_set_value(lcd_reset,1);
		mdelay(120);
#endif
		/* LCD needs SPI */

#ifdef DEBUG_INFO
		printk(KERN_INFO"%s",__func__);
#endif
		
		spi_init();

		ili9341_disp_powerup();

		ili9341_disp_on();

		ili9341_state.disp_initialized = TRUE;
	}
	return 0;
}

static int lcdc_ili9341_panel_off(struct platform_device *pdev)
{
	if (ili9341_state.disp_powered_up && ili9341_state.display_on) {
		/*serigo(GORDON_REG_LCDIFCTL2, 0x7B);
		serigo(GORDON_REG_VALTRAN, 0x01);
		serigo(GORDON_REG_LCDIFCTL1, 0x02);
		serigo(GORDON_REG_LCDIFCTL3, 0x01);
		mdelay(20);
		serigo(GORDON_REG_VALTRAN, 0x01);
		serigo(GORDON_REG_IMGCTL1, 0x01);
		serigo(GORDON_REG_LCDIFCTL1, 0x00);
		mdelay(20);

		serigo(GORDON_REG_POWCTL, 0x1F);
		mdelay(40);

		serigo(GORDON_REG_POWCTL, 0x07);
		mdelay(40);

		serigo(GORDON_REG_POWCTL, 0x03);
		mdelay(40);

		serigo(GORDON_REG_POWCTL, 0x00);
		mdelay(40);*/
#if 0
		lcdc_ili9341_pdata->panel_config_gpio(0);
#endif
#ifdef DEBUG_INFO
		printk(KERN_INFO"%s",__func__);
#endif
		ili9341_state.display_on = FALSE;
		ili9341_state.disp_initialized = FALSE;
	}
	return 0;
}
static void lcdc_ili9341_set_backlight(struct msm_fb_data_type *mfd)
{
		int bl_level = mfd->bl_level;
		if (bl_level <= 1) {
			/* keep back light OFF */
			//serigo(GORDON_REG_LCDIFCTL2, 0x0B);
		//	udelay(15);
			//serigo(GORDON_REG_VALTRAN, 0x01);
		} else {
			/* keep back light ON */
	                ili9341_write_command(0x51);       //Set Brightness 
		        ili9341_write_data(0x90); // 0x90 higher brightness than LK bootloader
			udelay(15);
	                ili9341_write_command(0x53);       //Set Brightness 
		        ili9341_write_data(0x2C); 
		}
}
static int __devinit ili9341_probe(struct platform_device *pdev)
{
	if (pdev->id == 0) {
		lcdc_ili9341_pdata = pdev->dev.platform_data;
		return 0;
	}
	msm_fb_add_device(pdev);
		ili9341_state.disp_powered_up = TRUE;
		ili9341_state.display_on = TRUE;
		ili9341_state.disp_initialized = TRUE;
	
	return 0;
}

static struct platform_driver this_driver = {
	.probe  = ili9341_probe,
	.driver = {
		.name   = "lcdc_ili9341_qvga",
	},
};

static struct msm_fb_panel_data ili9341_panel_data = {
	.on = lcdc_ili9341_panel_on,
	.off = lcdc_ili9341_panel_off,
	.set_backlight = lcdc_ili9341_set_backlight,
};

static struct platform_device this_device = {
	.name   = "lcdc_ili9341_qvga",
	.id	= 1,
	.dev	= {
		.platform_data = &ili9341_panel_data,
	}
};

static int __init lcdc_ili9341_panel_init(void)
{
	int ret;
	struct msm_panel_info *pinfo;

//#ifdef CONFIG_FB_MSM_LCDC_AUTO_DETECT
//	if (msm_fb_detect_client("lcdc_ili9341_qvga"))
//		return 0;
//#endif

	ret = platform_driver_register(&this_driver);
	if (ret)
		return ret;

	pinfo = &ili9341_panel_data.panel_info;
	pinfo->xres = 240;
	pinfo->yres = 320;
	MSM_FB_SINGLE_MODE_PANEL(pinfo);
	pinfo->type = LCDC_PANEL;
	pinfo->pdest = DISPLAY_1;
	pinfo->wait_cycle = 0;
	pinfo->bpp = 16;//18; try with 16 first, same as Inventec code
	pinfo->fb_num = 2;
	pinfo->clk_rate = 6000000;
	pinfo->bl_max = 0xFF;
	pinfo->bl_min = 0x01;

	/* Following six values are "typical" values taken from ILI9341 datasheet
	 * and may need some change later.
	 */
	pinfo->lcdc.h_back_porch = 20;   // Horizontal Back Porch
	pinfo->lcdc.h_front_porch = 10;  // Horizontal Front Porch
	pinfo->lcdc.h_pulse_width = 10;  // Horizontal Synchronization (Hsync)
	pinfo->lcdc.v_back_porch = 2;    // Vertical Back Porch
	pinfo->lcdc.v_front_porch = 4;   // Vertical Front Porch
	pinfo->lcdc.v_pulse_width = 2;   // Vertical Synchronization (Vsync)

	pinfo->lcdc.border_clr = 0;      /* blk */
	pinfo->lcdc.underflow_clr = 0xff;/* blue */
	pinfo->lcdc.hsync_skew = 0;

	ret = platform_device_register(&this_device);
	if (ret)
		platform_driver_unregister(&this_driver);

	return ret;
}

module_init(lcdc_ili9341_panel_init);
